const string info = "To give feedback about the campaigns, individidual maps or plugin, join SBVille's Discord:  \n" +
              Icons::Discord + " [Sbv's Lounge](https://discord.gg/HC3UfXJr3m)";

array<Campaign@>@ campaigns = {};
bool campaigns_fetched = false;

array<Campaign@>@ ParseCampaignsJson(Json::Value@ campaigns_json) {
    array<Campaign@>@ parsed = {};

    if (campaigns_json.GetType() != Json::Type::Array) {
        warn('Received campaign json not an array!');
        return parsed;
    }

    for (uint ci=0; ci<campaigns_json.get_Length(); ci++) {
        Json::Value@ campaign_json = campaigns_json[ci];
        
        if (campaign_json.GetType() != Json::Type::Object) {
            warn('Received campaign #' + ci + ' json is not an object!');
            continue;
        }

        try {
            string name = campaign_json.Get('name');

            string description = campaign_json.Get('description');

            Json::Value@ medal_json = campaign_json.Get('medal');
            Medal@ medal = Medal(medal_json.Get('on'), medal_json.Get('off'));

            Json::Value@ trophy_json = campaign_json.Get('trophy');
            Trophy@ trophy = null;
            if (trophy_json.GetType() == Json::Type::Object) {
                @trophy = Trophy(
                    trophy_json.Get('filename'),
                    trophy_json.Get('width'),
                    trophy_json.Get('height')
                );
            }

            Json::Value@ maps_json = campaign_json.Get('maps');
            array<ChallengeMap@> maps = {};
            for (uint mi=0; mi<maps_json.get_Length(); mi++) {
                Json::Value@ map_json = maps_json[mi];
                maps.InsertLast(ChallengeMap(
                    map_json.Get('uid'),
                    map_json.Get('name'),
                    map_json.Get('time'),
                    map_json.Get('color')
                ));
            }

            parsed.InsertLast(Campaign(
                name,
                description,
                medal,
                trophy,
                maps
            ));
        } catch {
            error(getExceptionInfo());
            warn('Failed to parse campaign #' + ci);
            continue;
        }
    }

    return parsed;
}

void LoadCampaignsFromOpenPlanetConfigs() {
    print('Fetching maps from OpenPlanet config.');
    Net::HttpRequest@ req = Net::HttpGet('https://openplanet.dev/plugin/sbvillecampaignchallenges/config/v1_4_0');

    uint maxWaits = 10;
    uint waits = 1;

    do {
        print('Waiting for OpenPlanet config (' + waits + '/' + maxWaits + ')');
        sleep(500);
        waits++;
    } while(!req.Finished() && waits <= maxWaits);

    if (req.Finished()) {
        int res_code = req.ResponseCode();
        if (res_code == 200) {
            campaigns_fetched = true;
            print('Fetch successful! Parsing...');
            Json::Value@ campaigns_json = req.Json();
            @campaigns = ParseCampaignsJson(campaigns_json);
            print('Campaigns parsed.');
        } else {
            warn('Campaigns fetch OpenPlanet.dev configs failed with response code ' + res_code);
        }
    } else {
        warn('Campaigns fetch from OpenPlanet.dev timed out!');
    }
}

Campaign@ findCampaignByMapUid(const string &in uid) {
    for (uint c=0; c<campaigns.Length; c++) {
        Campaign@ campaign = campaigns[c];
        for (uint m=0; m<campaign.maps.Length; m++) {
            ChallengeMap@ cm = campaign.maps[m];
            if (cm.uid == uid) {
                return campaign;
            }
        }
    }
    return null;
}

ChallengeMap@ findChallengeByUid(Campaign@ campaign, const string &in uid) {
    for (uint i=0; i<campaign.maps.Length; i++) {
        ChallengeMap@ cm = campaign.maps[i];
        if (cm.uid == uid) {
            return cm;
        }
    }
    return null;
}

CGameCtnChallengeInfo@ currentMap = null;
Campaign@ currentCampaign = null;
ChallengeMap@ currentChallenge = null;
bool showAll = false;
bool pbsLoading = false;
bool pbsLoadError = false;
bool showTrophy = false;
bool allowShowTrophy = false;

void OnCampaignChanged(Campaign@ campaign) {
    @currentCampaign = campaign;
    @currentChallenge = null;
    showAll = false;
    pbsLoading = false;
    pbsLoadError = false;
    showTrophy = false;

    if (currentCampaign is null) {
        return;
    }

    print('SBV Campaign changed to: ' + campaign.name);
    FetchAllPBs(campaign);
    allowShowTrophy = currentCampaign.trophy !is null && campaign.completed();
}

void OnChallengeChanged(ChallengeMap@ challenge) {
    @currentChallenge = challenge;

    if (currentChallenge is null) {
        return;
    }

    print('SBV Challenge changed to: ' + currentChallenge.name);
}

void OnMapChanged(CGameCtnChallengeInfo@ map) {
    @currentMap = map;
    
    if (currentMap is null) {
        @currentMap = null;
        @currentChallenge = null;
        return;
    }

    if (!campaigns_fetched) {
        LoadCampaignsFromOpenPlanetConfigs();
    }

    string mapUID = currentMap.MapUid;
    Campaign@ newCampaign = findCampaignByMapUid(mapUID);

    if (
        (newCampaign is null && currentCampaign !is null)
        || (currentCampaign is null && newCampaign !is null)
        || (currentCampaign !is null && newCampaign !is null && currentCampaign.name != newCampaign.name)
    ) {
        OnCampaignChanged(newCampaign);
    }

    if (currentCampaign is null) {
        return;
    }

    ChallengeMap@ newChallenge = findChallengeByUid(currentCampaign, mapUID);
    if (
        (newChallenge is null && currentChallenge !is null)
        || (currentChallenge is null && newChallenge !is null)
        || (currentChallenge !is null && newChallenge !is null && currentChallenge.uid != newChallenge.uid)
    ) {
        OnChallengeChanged(newChallenge);
    }
}

CTrackMania@ app;
CTrackManiaNetwork@ network;

void Main() {
    print('SBV Challenge plugin started');

    @app = cast<CTrackMania>(GetApp());
    @network = cast<CTrackManiaNetwork>(app.Network);

    LoadCampaignsFromOpenPlanetConfigs();

    while (true) {
        Tick();
        sleep(500);
    }
}

bool refreshPBs = false;

void Tick() {
    CGameCtnChallengeInfo@ newMap = null;
    if (
        app.RootMap !is null
        && app.RootMap.MapInfo !is null
        && app.RootMap.MapInfo.MapUid != ""
        && app.Editor is null
    ) {
        @newMap = app.RootMap.MapInfo;
    }

    if (
        (newMap is null && currentMap !is null)
        || (currentMap is null && newMap !is null)
        || (currentMap !is null && newMap !is null && currentMap.MapUid != newMap.MapUid)
    ) {
        OnMapChanged(newMap);
    }

    if (currentMap is null || currentChallenge is null || network.ClientManiaAppPlayground is null) {
        return;
    }

    auto userMgr = network.ClientManiaAppPlayground.UserMgr;
    if (userMgr.Users.Length == 0) {
        return;
    }

    auto scoreMgr = network.ClientManiaAppPlayground.ScoreMgr;
    uint newPB = scoreMgr.Map_GetRecord_v2(userMgr.Users[0].Id, currentChallenge.uid, "PersonalBest", "", "TimeAttack", "");
    uint oldPB = currentChallenge.pb;
    currentChallenge.setPB(newPB);
    if (
        currentChallenge.pb < oldPB && // this run was a new PB
        oldPB > currentChallenge.time && // old PB was slower than the challenge time
        currentChallenge.achieved() // but we got it now
    ) {
        // it's a new challenge achievement!
        // check if we got all now and show trophy immediately if so
        FetchAllPBs(currentCampaign);
        if (currentCampaign.completed() && currentCampaign.trophy !is null) {
            showTrophy = true;
            allowShowTrophy = true;
        }
    }
    
    if (refreshPBs) {
        print('Manually refreshing PBs');
        refreshPBs = false;
        FetchAllPBs(currentCampaign);
    }
}

const string tm_maps_endpoint = "https://live-services.trackmania.nadeo.live/api/token/leaderboard/group/map";

void FetchAllPBs(Campaign@ campaign) {
    pbsLoadError = false;
    pbsLoading = true;
    //try {
        if (!NadeoServices::IsAuthenticated("NadeoLiveServices")) {
            NadeoServices::AddAudience("NadeoLiveServices");
            uint maxAuthWait = 50;

            do {
                sleep(100);
                maxAuthWait--;
            } while (!NadeoServices::IsAuthenticated("NadeoLiveServices") && maxAuthWait > 0);

            if (!NadeoServices::IsAuthenticated("NadeoLiveServices")) {
                error('NadeoLiveServices auth failed');
                throw('auth fail');
            }
        }

        // Form request body
        Json::Value req_body = Json::Object();
        req_body["maps"] = Json::Array();
        for (uint i=0; i<campaign.maps.Length; i++) {
            ChallengeMap@ cm = campaign.maps[i];
            Json::Value reqmap = Json::Object();
            reqmap['groupUid'] = 'Personal_Best';
            reqmap['mapUid'] = cm.uid;
            req_body["maps"].Add(reqmap);
        }

        // Request and wait for response
        Net::HttpRequest@ req = NadeoServices::Post("NadeoLiveServices", tm_maps_endpoint, Json::Write(req_body));
        req.Start();
        uint maxPBWait = 100;

        do {
            sleep(100);
            maxPBWait--;
        } while(!req.Finished() && maxPBWait > 0);

        if (!req.Finished()) {
            error('NadeoLiveServices timeout while fetching PBs');
            throw('timeout');
        }

        if (req.ResponseCode() != 200) {
            error('NadeoLiveServices returned ' + req.ResponseCode() + ' while fetching PBs');
            throw('response error');
        }

        Json::Value res = Json::Parse(req.String());
        for (uint i=0; i<res.Length; i++) {
            // Worst case O(n^2) but whatever, list is small
            Json::Value res_map = res[i];
            string uid = res_map["mapUid"];
            if (currentChallenge !is null && uid == currentChallenge.uid) {
                // we don't want to update current map pb because
                // LBs are slightly slower than local time so
                // may cause an older, slower time to be updated
                // when fetched right after new PB
                continue;
            }
            ChallengeMap@ cm = findChallengeByUid(campaign, uid);
            if (cm !is null) {
                cm.setPB(res_map["score"]);
            }
        }
    /*} catch {
        error(getExceptionInfo());
        pbsLoadError = true;
    }*/
    pbsLoading = false;
}

const string color = '\\$0f0';
const string check = ' ' + Icons::Check;
            
void Render() {
    if (!UI::IsGameUIVisible()) return;

    if (currentChallenge is null) return;

    bool achieved = currentChallenge.achieved();
    
    int flags =
        UI::WindowFlags::NoTitleBar |
        UI::WindowFlags::AlwaysAutoResize |
        UI::WindowFlags::NoCollapse;
    UI::Begin("SBVille's Challenge", flags);

    UI::Image(achieved ? currentCampaign.medal.texture_on : currentCampaign.medal.texture_off, vec2(96, 96));

    UI::SameLine();
    UI::SetCursorPos(UI::GetCursorPos() + vec2(0, 20));
    UI::BeginGroup();
        UI::Text("\\$fa0SBVille's Challenge");
        string renderMapName = currentChallenge.name != '' ? ' - ' + currentChallenge.name : '';
        UI::Text(currentCampaign.name + renderMapName);
        UI::Text(
            (achieved ? color : '') + Icons::ClockO
            + ' ' + currentChallenge.timeFormatted + (achieved ? check : '')
            + '  ' + currentChallenge.deltaFormatted
        );
    UI::EndGroup();

    UI::NewLine();
    string detailsLabel = currentCampaign.maps.Length > 1
        ? 'Show all maps and info'
        : 'Shop map details and info';
    showAll = UI::Checkbox(detailsLabel, showAll);
    if (showAll) {
        UI::NewLine();
        if (pbsLoading) {
            UI::Text(Icons::Hourglass + ' Loading PBs...');
        } else if (pbsLoadError) {
            UI::Text('\\$f55Failed to load PBs. Try refresh later.');
        }

        if (currentCampaign.description.Length > 0) {
            UI::PushTextWrapPos(500.0);
            UI::TextWrapped(currentCampaign.description);
            UI::PopTextWrapPos();
            UI::NewLine();
        }

        if (UI::BeginTable('allmaps', 3, UI::TableFlags::SizingFixedFit)) {
            UI::TableNextRow();
            if (currentCampaign.maps.Length > 1) {
                UI::TableNextColumn();
                UI::Text('Map');
            }
            UI::TableNextColumn();
            UI::Text('Challenge');
            UI::TableNextColumn();
            UI::Text('Personal Best');

            for (uint i=0; i<currentCampaign.maps.Length; i++) {
                ChallengeMap@ cm = currentCampaign.maps[i];
                bool cm_achieved = cm.achieved();

                UI::TableNextRow();

                if (currentCampaign.maps.Length > 1) {
                    UI::TableNextColumn();
                    UI::Text('\\$' + cm.color + cm.name);
                }

                UI::TableNextColumn();
                UI::Text((cm_achieved ? color : '') + cm.timeFormatted);

                UI::TableNextColumn();
                if (cm.pbFormatted != '') {
                    UI::Text((cm_achieved ? color : '') + cm.pbFormatted + (cm_achieved ? check : '') + '  ' + cm.deltaFormatted);
                }
            }

            UI::EndTable();
        }
        
        if (UI::BeginTable('buttons', 2, UI::TableFlags::SizingFixedFit)) {
            UI::TableNextRow();
            UI::TableNextColumn();
            if (UI::Button('Refresh PB' + (currentCampaign.maps.Length > 1 ? 's' : ''))) {
                refreshPBs = true;
            }
            UI::TableNextColumn();
            if (allowShowTrophy && UI::Button('Show trophy')) {
                showTrophy = true;
            }
            UI::EndTable();
        }

        UI::NewLine();

        UI::Markdown(info);
    }

    UI::End();

    if (showTrophy) {
        int screenWidth = Draw::GetWidth();
        int screenHeight = Draw::GetHeight();
        float maxWidth = screenWidth * 0.85;
        float maxHeight = screenHeight * 0.85;
        float scaleX = maxWidth / currentCampaign.trophy.width;
        float scaleY = maxHeight / currentCampaign.trophy.height;
        float scale = (scaleX < scaleY) ? scaleX : scaleY;
        vec2 scaledSize = vec2(currentCampaign.trophy.width * scale, currentCampaign.trophy.height * scale);

        int posX = (screenWidth - scaledSize.x) * 0.5;
        int posY = (screenHeight - scaledSize.y) * 0.5;

        UI::SetNextWindowPos(posX, posY);
        UI::PushStyleColor(UI::Col::WindowBg, vec4(0, 0, 0, 0));

        int trophyWindowflags =
            UI::WindowFlags::NoTitleBar |
            UI::WindowFlags::AlwaysAutoResize |
            UI::WindowFlags::NoCollapse;
        UI::Begin('trophy', trophyWindowflags);

        UI::Image(currentCampaign.trophy.texture, scaledSize);
        UI::SetCursorPos(vec2(scaledSize.x * 0.5, scaledSize.y));
        if (UI::Button('Close', vec2(120, 50))) {
            showTrophy = false;
        }
        UI::End();

        UI::PopStyleColor();
    }
}